/**
 * Homework #2
 * CMSC 427 - Spring 2000.
 * 
 * @author Ben Bederson
 * @author Lance Good
 * @date February 11, 2000
 */
import java.awt.*;
import java.awt.event.*;

/**
 * The LineDrawing class is a placeholder to actually draw lines. It simply
 * creates a window with a drawing panel.
 */
public class CircleDrawing extends Frame
{
	// The number of colors to use in anti-aliasing
	public static final int NUM_ALIAS_COLORS = 16;
	
	// The actual pre-allocated array of colors used for anti-aliasing
	public static Color[] aliasColors = new Color[NUM_ALIAS_COLORS];
	
	/**
	 * The default constructor
	 */
    public CircleDrawing() {
		setLayout(new BorderLayout());
		setBounds(100, 100, 400, 400);           // Define the window dimensions
		DrawingPanel panel = new DrawingPanel(); // Create the drawing panel
		panel.setBounds(0, 0, 400, 400);         // Set the drawing panel size
		add(panel);                              // Add the drawing panel to the window
		
		Checkbox cb = new Checkbox("Anti-aliased",false); // A check box to specify anti-aliasing
		cb.addItemListener(new AliasItemListener(panel)); // Add an item listener to the box
		add(cb,"South");								  // Add the check box to the frame
		
		// Allocate the array of colors used for anti-aliasing
		int step = 256 / NUM_ALIAS_COLORS;
		for(int i=NUM_ALIAS_COLORS-1; i>=0; i--) {
			aliasColors[NUM_ALIAS_COLORS-1-i] = new Color(i*step,i*step,i*step);	
		}
		
		setVisible(true);                        // Make the whole thing visible

		// We must define this event handler
		// to capture clicks on the "X" to
		// exit the program.
		addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
    }

    /**
     * This gets called when the application starts up. The first thing
     * we do is create an instance of the main class.
     */
    public static void main(String[] args)
    {
		new CircleDrawing();
    }
}

/**
 * This is the drawing panel that we use to draw on. We must create this class
 * extending Panel so that we can override the crucial 'paint' method which
 * gets called whenever the system determines that Panel needs to be redrawn.
 * We also define simple event handlers to determine when the user clicks
 * on the drawing panel.
 */
class DrawingPanel extends Panel
{
	private int x1, y1, x2, y2;  // Center and radius measure for this circle
	private boolean antiAliased; // Is this panel anti-aliased
   
    /**
     * Simple constructor that adds a mouse listener to this panel.
     */
    public DrawingPanel() {
		DrawingMouseListener mouseListener = new DrawingMouseListener(this);
		addMouseListener(mouseListener);
		antiAliased = false;
    }

	/**
	 * Draws a circle from the given center point and point on the radius
	 * @param g Graphics for this panel
	 * @param xs The x coord for the center of the circle
	 * @param ys The y coord for the center of the circle
	 * @param xf The x coord to specify the radius of the circle
	 * @param yf The y coord to specify the radius of the circle
	 */
	public void drawCircle(Graphics g, int xs, int ys, int xf, int yf) {
		float radius = (float)Math.sqrt((xs-xf)*(xs-xf)+(ys-yf)*(ys-yf));
		int rndRadius = (int)(radius+0.5f);
		
		// Handle the small cases first outside of the loops.  This
		// increases performance by reducing special case checks in
		// the loop
		if (radius == 0.0f) {
		    g.setColor(CircleDrawing.aliasColors[CircleDrawing.NUM_ALIAS_COLORS-1]);
		    g.drawLine(xs,ys,xs,ys);
			return;
		}
		else if (radius < 2.0f && antiAliased) {
		    g.setColor(CircleDrawing.aliasColors[CircleDrawing.NUM_ALIAS_COLORS-1]);		    
		    g.drawLine(xs+1,ys,xs+1,ys);
		    g.drawLine(xs,ys+1,xs,ys+1);
		    g.drawLine(xs,ys-1,xs,ys-1);
		    g.drawLine(xs-1,ys,xs-1,ys);
		    g.setColor(CircleDrawing.aliasColors[CircleDrawing.NUM_ALIAS_COLORS/2]);
		    g.drawLine(xs+1,ys+1,xs+1,ys+1);
		    g.drawLine(xs+1,ys-1,xs+1,ys-1);
		    g.drawLine(xs-1,ys+1,xs-1,ys+1);
		    g.drawLine(xs-1,ys-1,xs-1,ys-1);
		    return;
		}

		// Now render the circle appropriately depending on
		// whether anti-aliased or not
		if (antiAliased) {
			int x = xs;
			// We will mirror our drawing in all 8 octants so determine
			// when x leaves the first octant and set stop equal to that
			int stop = (int)(Math.sqrt(radius*radius/2.0f)+0.5);
			while (x-xs <= stop) {
			    // Compute the actual circle coordinate at x
				float actual = (float)Math.sqrt(radius*radius-(float)(x-xs)*(x-xs));
				
				// Now compute the length of the intersection between the
				// line x'=x and the one pixel thick line tangent to
				// the circle at the point (x,actual)
				float segLength = 0.5f/(float)Math.cos(Math.atan((float)(x-xs)/actual));

				// Do a little math and rounding to find the top and bottom
				// coords of the previously mentioned intersection at x
				int top = (int)(actual + 1.0f + segLength) - 1;
				int bottom = (int)(actual - segLength);

				// Now we can calculate the color for the top and bottom
				// by the length of the segment of intersection in the
				// top and bottom pixels
				float topColor = (float)Math.abs(actual + segLength - (float)top);
				float bottomColor = (float)Math.abs((float)bottom + 1.0f - (actual - segLength));
				
				// Now actually draw the top pixel in all 8 octants
				if (topColor < 1) {
				    g.setColor(CircleDrawing.aliasColors[(int)(topColor*(float)CircleDrawing.NUM_ALIAS_COLORS)]);
				}
				else {
				    g.setColor(CircleDrawing.aliasColors[CircleDrawing.NUM_ALIAS_COLORS-1]);
				}
				g.drawLine(x,ys+top,x,ys+top);
				g.drawLine(x,ys-top,x,ys-top);
				g.drawLine(xs-(x-xs),ys+top,xs-(x-xs),ys+top);
				g.drawLine(xs-(x-xs),ys-top,xs-(x-xs),ys-top);
				g.drawLine(xs+top,ys+(x-xs),xs+top,ys+(x-xs));
				g.drawLine(xs+top,ys-(x-xs),xs+top,ys-(x-xs));
				g.drawLine(xs-top,ys+(x-xs),xs-top,ys+(x-xs));
				g.drawLine(xs-top,ys-(x-xs),xs-top,ys-(x-xs));

				// Now actually draw the bottom pixel in all 8 octants
				if (bottomColor < 1) {
				    g.setColor(CircleDrawing.aliasColors[(int)(bottomColor*(float)CircleDrawing.NUM_ALIAS_COLORS)]);
				}
				else {
				    g.setColor(CircleDrawing.aliasColors[CircleDrawing.NUM_ALIAS_COLORS-1]);
				}
				g.drawLine(x,ys+bottom,x,ys+bottom);	
				g.drawLine(x,ys-bottom,x,ys-bottom);
				g.drawLine(xs-(x-xs),ys+bottom,xs-(x-xs),ys+bottom);
				g.drawLine(xs-(x-xs),ys-bottom,xs-(x-xs),ys-bottom);
				g.drawLine(xs+bottom,ys+(x-xs),xs+bottom,ys+(x-xs));
				g.drawLine(xs+bottom,ys-(x-xs),xs+bottom,ys-(x-xs));
				g.drawLine(xs-bottom,ys+(x-xs),xs-bottom,ys+(x-xs));
				g.drawLine(xs-bottom,ys-(x-xs),xs-bottom,ys-(x-xs));
				
				// Draw any pixels in-between the top and bottom coords at x
				// using the darkest color available
				g.setColor(CircleDrawing.aliasColors[CircleDrawing.NUM_ALIAS_COLORS-1]);				
				for(int i=top-1; i>bottom; i--) {
					g.drawLine(x,ys+i,x,ys+i);
					g.drawLine(x,ys-i,x,ys-i);
					g.drawLine(xs-(x-xs),ys+i,xs-(x-xs),ys+i);
					g.drawLine(xs-(x-xs),ys-i,xs-(x-xs),ys-i);
					g.drawLine(xs+i,ys+(x-xs),xs+i,ys+(x-xs));
					g.drawLine(xs+i,ys-(x-xs),xs+i,ys-(x-xs));
					g.drawLine(xs-i,ys+(x-xs),xs-i,ys+(x-xs));
					g.drawLine(xs-i,ys-(x-xs),xs-i,ys-(x-xs));
				}
				
				// Now increment x
				x++;
			}
		}
		else {
			// This is the exact algorithm Ben gave in class
			int x = xs;
			int y = ys+rndRadius;
			int d =	5-4*rndRadius;
			while(x-xs <= y-ys) {
				// Draw the pixel in all 8 octants
				g.drawLine(x,y,x,y);
				g.drawLine(x,ys-(y-ys),x,ys-(y-ys));
				g.drawLine(xs-(x-xs),y,xs-(x-xs),y);
				g.drawLine(xs-(x-xs),ys-(y-ys),xs-(x-xs),ys-(y-ys));
				g.drawLine(xs+(y-ys),ys+(x-xs),xs+(y-ys),ys+(x-xs));
				g.drawLine(xs+(y-ys),ys-(x-xs),xs+(y-ys),ys-(x-xs));
				g.drawLine(xs-(y-ys),ys+(x-xs),xs-(y-ys),ys+(x-xs));
				g.drawLine(xs-(y-ys),ys-(x-xs),xs-(y-ys),ys-(x-xs));
				
				if (d < 0) {
					d = d + 8*(x-xs) + 12;
					x++;
				}
				else {
					d = d + 8*(x-xs) - 8*(y-ys) + 20;
					x++;
					y--;
				}
			}
		}		
	}
		
    /**
     * This method gets called whenever the panel needs to get repainted. 
     * Putting in drawing methods here, such as g.drawLine() will cause
     * those things to be drawn on the panel. For instance, if you add
     * g.drawLine(0, 0, 100, 100) the canvas will appear with a diagonal line.
     */
    public void paint(Graphics g) {		
	drawCircle(g,x1,y1,x2,y2);
    }

    /**
     * Specify the first point for the center of the circle
     * @param x The x coord of the point
     * @param y The y coord of the point
     */
    public void setPoint1(int x, int y) {
		x1 = x;
		y1 = y;
		repaint();
    }

    /**
     * Specify the second point to determine the radius of the circle
     * @param x The x coord of the point
     * @param y The y coord of the point
     */
    public void setPoint2(int x, int y) {
		x2 = x;
		y2 = y;
		repaint();
    }
	
	public void setAntiAliasing(boolean antiAlias) {
		if (antiAlias != this.antiAliased) {
			this.antiAliased = antiAlias;
			repaint();
		}
	}
}

/**
 * A mouse listener for a DrawingPanel that passes the first and second
 * point of a line back to the panel with alternating mouse clicks.
 */
class DrawingMouseListener extends MouseAdapter {
    private DrawingPanel panel; // The panel to draw on
    private boolean firstPress = true; // True when expecting first mouse press

    /**
     * Constructor that receives the panel to draw on.
     */
    public DrawingMouseListener(DrawingPanel panel) {
		this.panel = panel;
    }

    /**
     * Mouse Press event listener that sets both points of the line on 
     * the first mouse click, and then sets the second point on the
     * second click - effectively letting the user draw a line one
     * with each two clicks.
     */
    public void mousePressed(MouseEvent e) {
		if (firstPress) {
		    panel.setPoint1(e.getX(), e.getY());
		    panel.setPoint2(e.getX(), e.getY());
			firstPress = false;
		} else {
		    panel.setPoint2(e.getX(), e.getY());
			firstPress = true;
		}
    }
		
}

/**
 * This is a simple item listener to listen for changes in the state of
 * the checkbox used to turn anti-aliasing on and off
 */
class AliasItemListener implements ItemListener {
	private DrawingPanel panel;
	
	/**
	 * The default constructor
	 * @param panel The drawing panel on which anti-aliasing is turned on and off
	 */
	public AliasItemListener(DrawingPanel panel) {
		this.panel = panel;
	}
	
	/**
	 * Listens for item state changed events and turns anti-aliasing on and off
	 * @param e The actual item event
	 */
	public void itemStateChanged(ItemEvent e) {
		panel.setAntiAliasing(e.getStateChange() == ItemEvent.SELECTED);
	}
}
